/*

Copyright (C) 2001, James B. Rawlings

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.

*/

#include <math.h>
#include <stdio.h>

#include <mex.h>

static double
get_rand (void)
{
  double rnd;

  mxArray *plhs[1];

  mexCallMATLAB (1, plhs, 0, 0, "rand");

  rnd = (mxGetPr (plhs[0]))[0];

  mxDestroyArray (plhs[0]);

  return rnd;
}

void
mexFunction (int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
  int nargin = nrhs;

  const mxArray *xin;    double *pxin;
  const mxArray *k;      double *pk;
  const mxArray *stoiT;  double *pstoiT;  int stoiT_nr;
  const mxArray *tout;   double *ptout;

  mxArray *xout;   double *pxout;

  int nts, nx;
  double tfin, time;
  int iout = 0;
  double x0, x1, x2;
  double k0, k1, k2, k3, k4, k5;
  int pxout_offset = 0;

  if (nargin < 4 || nargin > 4)
    mexErrMsgTxt ("usage: stochsim (x0, k, stoiT, tout)");

  /* Maybe we should check that these are numeric, to avoid crashing
     Matlab if someone passes something else?  */
  xin = prhs[0];
  k = prhs[1];
  stoiT = prhs[2];
  tout = prhs[3];

  pxin = mxGetPr (xin);
  pk = mxGetPr (k);
  pstoiT = mxGetPr (stoiT);
  ptout = mxGetPr (tout);

  stoiT_nr = mxGetM (stoiT);

  nts = mxGetNumberOfElements (tout);
  tfin = ptout[nts-1];

  time = ptout[0];

  nx = mxGetNumberOfElements (xin);

  if (nx != 3)
    mexErrMsgTxt ("stochsim: expecting X to be a 3-element vector");

  x0 = pxin[0];
  x1 = pxin[1];
  x2 = pxin[2];

  xout = mxCreateDoubleMatrix (nx, nts, mxREAL);
  pxout = mxGetPr (xout);

  k0 = pk[0];
  k1 = pk[1];
  k2 = pk[2];
  k3 = pk[3];
  k4 = pk[4];
  k5 = pk[5];

  while (time < tfin) 
    {
      double r0 = k0 * x0;
      double r1 = k1 * x1;
      double r2 = k2 * x0;
      double r3 = k3 * x0;
      double r4 = k4 * x2;
      double r5 = k5 * x1 * x2;

      double rtot = r0 + r1 + r2 + r3 + r4 + r5;

      double timenew, rnd, rcum, tmp;
      int m = 0;
      double xnew0, xnew1, xnew2;
      
      /* return if system has extinguished */
      if (rtot == 0.0)
	{
	  while (iout < nts)
	    {
	      pxout[pxout_offset++] = x0;
	      pxout[pxout_offset++] = x1;
	      pxout[pxout_offset++] = x2;

	      iout++;
	    }
	  break;
	}

      /* choose likely time of next reaction */
      timenew = time - log (get_rand ()) / rtot;

      /* choose which reaction (mth) is likely to occur */
      rnd = get_rand ();
      rcum = r0;
      m = 0;
      tmp = rnd*rtot;
      if (rcum <= tmp)
	{
	  ++m; rcum += r1;

	  if (rcum <= tmp)
	    {
	      ++m; rcum += r2;

	      if (rcum <= tmp)
		{
		  ++m; rcum += r3;

		  if (rcum <= tmp)
		    {
		      ++m; rcum += r4;

		      if (rcum <= tmp)
			++m; rcum += r5;
		    }
		}
	    }
	}

      xnew0 = x0 + pstoiT[stoiT_nr*m];
      xnew1 = x1 + pstoiT[stoiT_nr*m+1];
      xnew2 = x2 + pstoiT[stoiT_nr*m+2];

      /* check if passing time of requested output */
      while (ptout[iout] <= timenew)
	{
	  pxout[pxout_offset++] = x0;
	  pxout[pxout_offset++] = x1;
	  pxout[pxout_offset++] = x2;

	  iout++;

	  if (iout >= nts)
	    break;
	}

      time = timenew;

      x0 = xnew0;
      x1 = xnew1;
      x2 = xnew2;
    }

  plhs[0] = xout;
}
